@@ -5,19 +5,31 @@
  *
  */
 #include "quickphrase.h"
-#include <fcntl.h>
 
+#include <array>
+#include <memory>
+#include <string>
 #include <utility>
 #include "fcitx-config/iniparser.h"
-#include "fcitx-utils/charutils.h"
+#include "fcitx-utils/capabilityflags.h"
+#include "fcitx-utils/handlertable.h"
 #include "fcitx-utils/i18n.h"
 #include "fcitx-utils/inputbuffer.h"
-#include "fcitx-utils/standardpath.h"
-#include "fcitx-utils/utf8.h"
+#include "fcitx-utils/key.h"
+#include "fcitx-utils/keysym.h"
+#include "fcitx-utils/textformatflags.h"
+#include "fcitx/addonfactory.h"
+#include "fcitx/addoninstance.h"
 #include "fcitx/addonmanager.h"
 #include "fcitx/candidatelist.h"
+#include "fcitx/event.h"
 #include "fcitx/inputcontextmanager.h"
 #include "fcitx/inputpanel.h"
+#include "fcitx/instance.h"
+#include "fcitx/text.h"
+#include "fcitx/userinterface.h"
+#include "quickphrase_public.h"
+#include "quickphraseprovider.h"
 
 namespace fcitx {
 
@@ -121,7 +133,7 @@ QuickPhrase::QuickPhrase(Instance *instance)
                 }
 
                 if (keyEvent.key().check(FcitxKey_space) &&
-                    candidateList->size()) {
+                    !candidateList->empty()) {
                     keyEvent.accept();
                     if (candidateList->cursorIndex() >= 0) {
                         candidateList->candidate(candidateList->cursorIndex())
@@ -164,7 +176,7 @@ QuickPhrase::QuickPhrase(Instance *instance)
                     }
                 }
 
-                if (candidateList->size() &&
+                if (!candidateList->empty() &&
                     keyEvent.key().checkKeyList(
                         instance_->globalConfig().defaultPrevCandidate())) {
                     keyEvent.filterAndAccept();
@@ -174,7 +186,7 @@ QuickPhrase::QuickPhrase(Instance *instance)
                     return;
                 }
 
-                if (candidateList->size() &&
+                if (!candidateList->empty() &&
                     keyEvent.key().checkKeyList(
                         instance_->globalConfig().defaultNextCandidate())) {
                     keyEvent.filterAndAccept();
@@ -245,21 +257,21 @@ QuickPhrase::QuickPhrase(Instance *instance)
                     state->buffer_.setCursor(0);
                     keyEvent.accept();
                     return updateUI(inputContext);
-                } else if (key.check(FcitxKey_End) ||
-                           key.check(FcitxKey_KP_End)) {
+                }
+                if (key.check(FcitxKey_End) || key.check(FcitxKey_KP_End)) {
                     state->buffer_.setCursor(state->buffer_.size());
                     keyEvent.accept();
                     return updateUI(inputContext);
-                } else if (key.check(FcitxKey_Left) ||
-                           key.check(FcitxKey_KP_Left)) {
+                }
+                if (key.check(FcitxKey_Left) || key.check(FcitxKey_KP_Left)) {
                     auto cursor = state->buffer_.cursor();
                     if (cursor > 0) {
                         state->buffer_.setCursor(cursor - 1);
                     }
                     keyEvent.accept();
                     return updateUI(inputContext);
-                } else if (key.check(FcitxKey_Right) ||
-                           key.check(FcitxKey_KP_Right)) {
+                }
+                if (key.check(FcitxKey_Right) || key.check(FcitxKey_KP_Right)) {
                     auto cursor = state->buffer_.cursor();
                     if (cursor < state->buffer_.size()) {
                         state->buffer_.setCursor(cursor + 1);
@@ -329,9 +341,12 @@ class QuickPhraseCandidateWord : public CandidateWord {
 public:
     QuickPhraseCandidateWord(QuickPhrase *q, std::string commit,
                              const std::string &display,
+                             const std::string &comment,
                              QuickPhraseAction action)
         : CandidateWord(Text(display)), q_(q), commit_(std::move(commit)),
-          action_(action) {}
+          action_(action) {
+        setComment(Text(comment));
+    }
 
     void select(InputContext *inputContext) const override {
         auto *state = inputContext->propertyFor(&q_->factory());
@@ -402,6 +417,8 @@ void QuickPhrase::updateUI(InputContext *inputContext) {
     inputContext->inputPanel().reset();
     if (!state->buffer_.empty()) {
         auto candidateList = std::make_unique<CommonCandidateList>();
+        candidateList->setCursorPositionAfterPaging(
+            CursorPositionAfterPaging::ResetToFirst);
         candidateList->setPageSize(instance_->globalConfig().defaultPageSize());
         QuickPhraseProvider *providers[] = {&callbackProvider_,
                                             &builtinProvider_, &spellProvider_};
@@ -413,9 +430,9 @@ void QuickPhrase::updateUI(InputContext *inputContext) {
             if (!provider->populate(
                     inputContext, state->buffer_.userInput(),
                     [this, &candidateList, &selectionKeyAction, &autoCommit,
-                     &autoCommitSet](const std::string &word,
-                                     const std::string &aux,
-                                     QuickPhraseAction action) {
+                     &autoCommitSet](
+                        const std::string &word, const std::string &aux,
+                        const std::string &comment, QuickPhraseAction action) {
                         if (!autoCommitSet &&
                             action == QuickPhraseAction::AutoCommit) {
                             autoCommit = word;
@@ -427,7 +444,7 @@ void QuickPhrase::updateUI(InputContext *inputContext) {
                         }
                         if (!word.empty()) {
                             candidateList->append<QuickPhraseCandidateWord>(
-                                this, word, aux, action);
+                                this, word, aux, comment, action);
                         } else {
                             if (action == QuickPhraseAction::DigitSelection ||
                                 action == QuickPhraseAction::AlphaSelection ||
@@ -486,11 +503,17 @@ void QuickPhrase::reloadConfig() {
     builtinProvider_.reloadConfig();
     readAsIni(config_, "conf/quickphrase.conf");
 }
+
 std::unique_ptr<HandlerTableEntry<QuickPhraseProviderCallback>>
 QuickPhrase::addProvider(QuickPhraseProviderCallback callback) {
     return callbackProvider_.addCallback(std::move(callback));
 }
 
+std::unique_ptr<HandlerTableEntry<QuickPhraseProviderCallbackV2>>
+QuickPhrase::addProviderV2(QuickPhraseProviderCallbackV2 callback) {
+    return callbackProvider_.addCallback(std::move(callback));
+}
+
 void QuickPhrase::trigger(InputContext *ic, const std::string &text,
                           const std::string &prefix, const std::string &str,
                           const std::string &alt, const Key &key) {
