@@ -6,19 +6,27 @@
  */
 #include "quickphraseprovider.h"
 #include <fcntl.h>
+#include <cstddef>
+#include <string>
+#include <string_view>
+#include <utility>
 #include <fcitx-utils/utf8.h>
 #include <fcitx/inputmethodentry.h>
 #include "fcitx-utils/fs.h"
+#include "fcitx-utils/macros.h"
+#include "fcitx-utils/misc.h"
 #include "fcitx-utils/standardpath.h"
 #include "fcitx-utils/stringutils.h"
+#include "fcitx/inputcontext.h"
 #include "quickphrase.h"
+#include "quickphrase_public.h"
 #include "spell_public.h"
 
 namespace fcitx {
 
 bool BuiltInQuickPhraseProvider::populate(
     InputContext *, const std::string &userInput,
-    const QuickPhraseAddCandidateCallback &addCandidate) {
+    const QuickPhraseAddCandidateCallbackV2 &addCandidate) {
     auto start = map_.lower_bound(userInput);
     auto end = map_.end();
 
@@ -26,9 +34,7 @@ bool BuiltInQuickPhraseProvider::populate(
         if (!stringutils::startsWith(start->first, userInput)) {
             break;
         }
-        addCandidate(start->second,
-                     stringutils::concat(start->second, " ",
-                                         start->first.substr(userInput.size())),
+        addCandidate(start->second, start->second, start->first,
                      QuickPhraseAction::Commit);
     }
     return true;
@@ -36,28 +42,28 @@ bool BuiltInQuickPhraseProvider::populate(
 void BuiltInQuickPhraseProvider::reloadConfig() {
 
     map_.clear();
-    auto file = StandardPath::global().open(StandardPath::Type::PkgData,
-                                            "data/QuickPhrase.mb", O_RDONLY);
-    auto files = StandardPath::global().multiOpen(
-        StandardPath::Type::PkgData, "data/quickphrase.d/", O_RDONLY,
-        filter::Suffix(".mb"));
-    auto disableFiles = StandardPath::global().multiOpen(
-        StandardPath::Type::PkgData, "data/quickphrase.d/", O_RDONLY,
-        filter::Suffix(".mb.disable"));
-    if (file.fd() >= 0) {
-        load(file);
+    if (auto file = StandardPath::global().open(
+            StandardPath::Type::PkgData, "data/QuickPhrase.mb", O_RDONLY);
+        file.fd() >= 0) {
+        load(fs::openFD(file, "rb"));
     }
 
+    auto files = StandardPath::global().locate(StandardPath::Type::PkgData,
+                                               "data/quickphrase.d/",
+                                               filter::Suffix(".mb"));
+    auto disableFiles = StandardPath::global().locate(
+        StandardPath::Type::PkgData, "data/quickphrase.d/",
+        filter::Suffix(".mb.disable"));
     for (auto &p : files) {
         if (disableFiles.count(stringutils::concat(p.first, ".disable"))) {
             continue;
         }
-        load(p.second);
+        UnixFD fd = UnixFD::own(open(p.second.c_str(), O_RDONLY));
+        load(fs::openFD(fd, "rb"));
     }
 }
 
-void BuiltInQuickPhraseProvider::load(StandardPathFile &file) {
-    UniqueFilePtr fp = fs::openFD(file, "rb");
+void BuiltInQuickPhraseProvider::load(UniqueFilePtr fp) {
     if (!fp) {
         return;
     }
@@ -65,13 +71,10 @@ void BuiltInQuickPhraseProvider::load(StandardPathFile &file) {
     UniqueCPtr<char> buf;
     size_t len = 0;
     while (getline(buf, &len, fp.get()) != -1) {
-        std::string strBuf(buf.get());
-
-        auto [start, end] = stringutils::trimInplace(strBuf);
-        if (start == end) {
+        std::string_view text = stringutils::trimView(buf.get());
+        if (text.empty()) {
             continue;
         }
-        std::string_view text(strBuf.data() + start, end - start);
         if (!utf8::validate(text)) {
             continue;
         }
@@ -87,8 +90,7 @@ void BuiltInQuickPhraseProvider::load(StandardPathFile &file) {
         }
 
         std::string key(text.begin(), text.begin() + pos);
-        auto wordString =
-            stringutils::unescapeForValue(std::string_view(text.substr(word)));
+        auto wordString = stringutils::unescapeForValue(text.substr(word));
 
         if (!wordString) {
             continue;
@@ -97,21 +99,21 @@ void BuiltInQuickPhraseProvider::load(StandardPathFile &file) {
     }
 }
 
-SpellQuickPhraseProvider::SpellQuickPhraseProvider(QuickPhrase *quickPhrase)
-    : parent_(quickPhrase), instance_(parent_->instance()) {}
+SpellQuickPhraseProvider::SpellQuickPhraseProvider(QuickPhrase *parent)
+    : parent_(parent), instance_(parent_->instance()) {}
 
 bool SpellQuickPhraseProvider::populate(
     InputContext *ic, const std::string &userInput,
-    const QuickPhraseAddCandidateCallback &addCandidate) {
+    const QuickPhraseAddCandidateCallbackV2 &addCandidate) {
     if (!*parent_->config().enableSpell) {
         return true;
     }
-    auto spell = this->spell();
+    auto *spell = this->spell();
     if (!spell) {
         return true;
     }
     std::string lang = *parent_->config().fallbackSpellLanguage;
-    if (auto entry = instance_->inputMethodEntry(ic)) {
+    if (const auto *entry = instance_->inputMethodEntry(ic)) {
         if (spell->call<ISpell::checkDict>(entry->languageCode())) {
             lang = entry->languageCode();
         } else if (!spell->call<ISpell::checkDict>(lang)) {
@@ -121,19 +123,29 @@ bool SpellQuickPhraseProvider::populate(
     const auto result = spell->call<ISpell::hint>(
         lang, userInput, instance_->globalConfig().defaultPageSize());
     for (const auto &word : result) {
-        addCandidate(word, word, QuickPhraseAction::Commit);
+        addCandidate(word, word, "", QuickPhraseAction::Commit);
     }
     return true;
 }
 
 bool CallbackQuickPhraseProvider::populate(
     InputContext *ic, const std::string &userInput,
-    const QuickPhraseAddCandidateCallback &addCandidate) {
-    for (const auto &callback : callback_.view()) {
+    const QuickPhraseAddCandidateCallbackV2 &addCandidate) {
+    for (const auto &callback : callbackV2_.view()) {
         if (!callback(ic, userInput, addCandidate)) {
             return false;
         }
     }
+    for (const auto &callback : callback_.view()) {
+        if (!callback(ic, userInput,
+                      [&addCandidate](const std::string &word,
+                                      const std::string &aux,
+                                      QuickPhraseAction action) {
+                          return addCandidate(word, aux, "", action);
+                      })) {
+            return false;
+        }
+    }
     return true;
 }
 
